/*
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that the
 * following conditions are met:
 *
 * 1. Redistributions of source code must retain this list
 *    of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce this
 *    list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include <config.h>
#include <stdlib.h>
#include <unistd.h>

#include "datpacker.h"

extern thdat_fptrs th06dat_fptrs;
extern thdat_fptrs th07dat_fptrs;
extern thdat_fptrs th08dat_fptrs;
extern thdat_fptrs th95dat_fptrs;

char library_error[LIBRARY_ERROR_SIZE] = { 0 };

static void
print_usage()
{
	puts("Usage: datpacker [OPTION]... ARCHIVE FILE...\n"
	     "OPTION can be:\n"
	     "  -h           prints the usage\n"
	     "  -v VERSION   can be 6, 7, 8, 9, 95, 10, or 11, defaults to 11\n"
		 "  -V           print program version\n\n"
		 "Report bugs to <" PACKAGE_BUGREPORT  ">.");
}

static void
print_version()
{
	puts("This is " PACKAGE_NAME " release " PACKAGE_VERSION ".");
}

int
main(int argc, char* argv[])
{
	FILE* archive;
	void* private;
	thdat_fptrs* fptrs = NULL;
	unsigned int version = 11;
	unsigned int i;
	int opt;

	opterr = 0;
	while ((opt = getopt(argc, argv, "hVv:")) != -1) {
		if (opt == 'h') {
			print_usage();
			return EXIT_SUCCESS;
		} else if (opt == 'v') {
			version = strtol(optarg, NULL, 10);
		} else if (opt == 'V') {
			print_version();
			return EXIT_SUCCESS;
		} else {
			print_usage();
			return EXIT_FAILURE;
		}
	}

	if ((optind + 2) > argc) {
		print_usage();
		return EXIT_FAILURE;
	}

	switch (version) {
	case 6:
		fptrs = &th06dat_fptrs;
		break;
	case 7:
		fptrs = &th07dat_fptrs;
		break;
	case 8:
	case 9:
		fptrs = &th08dat_fptrs;
		break;
	case 95:
	case 10:
	case 11:
		fptrs = &th95dat_fptrs;
		break;
	}

	if (!fptrs) {
		fprintf(stderr, "datpacker: unsupported version\n");
		return EXIT_FAILURE;
	}

	archive = fopen(argv[optind], "wb");
	if (!archive) {
		fprintf(stderr, "datpacker: couldn't open %s for writing\n", argv[optind]);
		return EXIT_FAILURE;
	}

	private = fptrs->open(archive, version);
	if (!private) {
		fprintf(stderr, "datpacker: %s\n", library_error);
		fclose(archive);
		return EXIT_FAILURE;
	}

	for (i = optind + 1; i < argc; ++i) {
		FILE* fd = fopen(argv[i], "rb");
		if (!fd) {
			fprintf(stderr, "datpacker: couldn't open %s for reading\n", argv[i]);
			fclose(archive);
			return EXIT_FAILURE;
		}
		printf("%s... ", argv[i]);
		fflush(stdout);
		if (fptrs->write(private, fd, argv[i]) == -1) {
			fprintf(stderr, "datpacker: %s\n", library_error);
			fclose(fd);
			fclose(archive);
			return EXIT_FAILURE;
		}
		printf("OK!\n");
		fclose(fd);
	}

	if (fptrs->close(private) == -1) {
		fprintf(stderr, "datpacker: %s\n", library_error);
		fclose(archive);
		return EXIT_FAILURE;
	}

	fclose(archive);

	return EXIT_SUCCESS;
}
