/*
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that the
 * following conditions are met:
 *
 * 1. Redistributions of source code must retain this list
 *    of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce this
 *    list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include <config.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <errno.h>

#include "datpacker.h"
#include "thlzss.h"
#include "util.h"
#include "bits.h"
#include "thdat.h"

static void
th06_write_uint(bitstream_t* b, uint32_t value)
{
	unsigned int size;
	if (value & 0xff000000) {
		size = 4;
	} else if (value & 0xff0000) {
		size = 3;
	} else if (value & 0xff00) {
		size = 2;
	} else {
		size = 1;
	}
	bitstream_write(b, 2, size - 1);
	bitstream_write(b, size * 8, value);
}

static void
th06_write_string(bitstream_t* b, unsigned int length, unsigned char* data)
{
	unsigned int i;
	for (i = 0; i < length; ++i) {
		bitstream_write(b, 8, data[i]);
	}
}

static archive_t*
th06_open(FILE* fd, unsigned int version)
{
	if (version != 6) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "internal error: wrong version passed: %u", version);
		return NULL;
	}

	return archive_open(fd, version, 13);
}

static int
th06_write(archive_t* archive, FILE* fd, const char* filename)
{
	unsigned int i;
	entry_t* entry;
	unsigned char* zdata;

	if (!archive) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "internal error: archive is NULL");
		return -1;
	}

	if (!filename) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "internal error: filename is NULL");
		archive_free(archive);
		return -1;
	}

	entry = thdat_write(archive, fd, filename);
	if (!entry) {
		archive_free(archive);
		return -1;
	}

	zdata = th_lz_fd(fd, &entry->zsize);
	if (!zdata) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "th_lz_mem failed (TODO: Improve error message.)");
		archive_free(archive);
		return -1;
	}

	for (i = 0; i < entry->zsize; ++i) {
		entry->extra += zdata[i];
	}

	if (fwrite(zdata, entry->zsize, 1, archive->fd) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't write: %s", strerror(errno));
		free(zdata);
		archive_free(archive);
		return -1;
	}
	free(zdata);

	archive->offset += entry->zsize;

	return 0;
}

static int
th06_close(archive_t* archive)
{
	const char* magic = "PBG3";
	unsigned int i;

	if (!archive) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "internal error: archive is NULL");
		return -1;
	}

	bitstream_t b;
	bitstream_init(&b, 16);

	/* Write entry directory. */
	for (i = 0; i < archive->count; ++i) {
		entry_t* entry = &archive->entries[i];
		/* These values are unknown, but it seems they can be ignored. */
		uint32_t unknown1 = 0; /* The same for all entries in an archive. */
		uint32_t unknown2 = 0; /* Starts at a high value.  Increases by a random multiple of a thousand per entry. */

		th06_write_uint(&b, unknown1);
		th06_write_uint(&b, unknown2);
		th06_write_uint(&b, entry->extra);
		th06_write_uint(&b, entry->offset);
		th06_write_uint(&b, entry->size);
		/* XXX: I hear something about a 0x100 max filename length.  Probably
		 * pretty unlikely to happen as most operating systems are limited to
		 * 0x100 long file names. */
		th06_write_string(&b, strlen(entry->name) + 1, (unsigned char*)entry->name);
	}

	bitstream_finish(&b);

	if (fwrite(b.buffer, b.used_bytes, 1, archive->fd) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't write: %s", strerror(errno));
		free(b.buffer);
		archive_free(archive);
		return -1;
	}

	free(b.buffer);

	if (fseek(archive->fd, 0, SEEK_SET) == -1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't seek: %s", strerror(errno));
		archive_free(archive);
		return -1;
	}

	if (fwrite(magic, 4, 1, archive->fd) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't write: %s", strerror(errno));
		archive_free(archive);
		return -1;
	}

	bitstream_init(&b, 8);

	th06_write_uint(&b, archive->count);
	th06_write_uint(&b, archive->offset);

	bitstream_finish(&b);

	if (fwrite(b.buffer, b.used_bytes, 1, archive->fd) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't write: %s", strerror(errno));
		free(b.buffer);
		archive_free(archive);
		return -1;
	}

	free(b.buffer);
	archive_free(archive);

	return 0;
}

thdat_fptrs th06dat_fptrs = {
	th06_open,
	th06_write,
	th06_close
};
