/*
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that the
 * following conditions are met:
 *
 * 1. Redistributions of source code must retain this list
 *    of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce this
 *    list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include <config.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <errno.h>

#include "datpacker.h"
#include "thlzss.h"
#include "util.h"
#include "thdat.h"

static archive_t*
th07_open(FILE* fd, unsigned int version)
{
	if (version != 7) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "internal error: wrong version passed: %u", version);
		return NULL;
	}

	return archive_open(fd, version, 16);
}

static int
th07_write(archive_t* archive, FILE* fd, const char* filename)
{
	entry_t* entry;
	unsigned char* data;
	unsigned char* zdata;

	if (!archive) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "internal error: archive is NULL");
		return -1;
	}

	if (!filename) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "internal error: filename is NULL");
		archive_free(archive);
		return -1;
	}

	entry = thdat_write(archive, fd, filename);
	if (!entry) {
		archive_free(archive);
		return -1;
	}

	data = malloc(entry->size);
	if (!data) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "malloc failed: %s", strerror(errno));
		archive_free(archive);
		return -1;
	}

	if (fread(data, entry->size, 1, fd) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't read: %s", strerror(errno));
		free(data);
		archive_free(archive);
		return -1;
	}

	zdata = th_lz_mem(data, entry->size, &entry->zsize);
	if (!zdata) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "th_lz_mem failed (TODO: Improve error message.)");
		archive_free(archive);
		return -1;
	}

	/* XXX: I do believe you can have files uncompressed data here and in th06. */

	if (fwrite(zdata, entry->zsize, 1, archive->fd) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't write: %s", strerror(errno));
		free(zdata);
		archive_free(archive);
		return -1;
	}
	free(zdata);

	archive->offset += entry->zsize;

	return 0;
}

static int
th07_close(archive_t* archive)
{
	unsigned char* buffer;
	unsigned char* buffer_ptr;
	unsigned int i;
	unsigned char* zbuffer;
	uint32_t header[4];
	unsigned int list_size = 0;
	unsigned int list_zsize = 0;

	if (!archive) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "archive is NULL");
		return -1;
	}

	for (i = 0; i < archive->count; ++i) {
		list_size += strlen(archive->entries[i].name) + 1 + (sizeof(uint32_t) * 3);
	}

	/* Write entry directory. */
	list_size += 4;
	buffer = malloc(list_size);
	if (!buffer) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "malloc failed: %s", strerror(errno));
		archive_free(archive);
		return -1;
	}
	memset(buffer, 0, list_size);

	buffer_ptr = buffer;
	for (i = 0; i < archive->count; ++i) {
		entry_t* entry = &archive->entries[i];
		buffer_ptr = write_bytes((unsigned char*)entry->name, strlen(entry->name) + 1, buffer_ptr);
		buffer_ptr = write_uint32(entry->offset, buffer_ptr);
		buffer_ptr = write_uint32(entry->size, buffer_ptr);
		buffer_ptr = write_uint32(0, buffer_ptr);
	}

	uint32_t zero = 0;
	memcpy(buffer_ptr, &zero, sizeof(uint32_t));

	zbuffer = th_lz_mem(buffer, list_size, &list_zsize);
	if (!zbuffer) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "th_lz_mem failed (TODO: Improve error message.)");
		free(buffer);
		archive_free(archive);
		return -1;
	}
	free(buffer);

	if (fwrite(zbuffer, list_zsize, 1, archive->fd) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't write: %s", strerror(errno));
		free(zbuffer);
		archive_free(archive);
		return -1;
	}
	free(zbuffer);

	/* Write header. */
	header[0] = 0x34474250; /* 4GBP */
	header[1] = archive->count;
	header[2] = archive->offset;
	header[3] = list_size;

	if (fseek(archive->fd, 0, SEEK_SET) == -1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't seek: %s", strerror(errno));
		archive_free(archive);
		return -1;
	}
	if (fwrite(&header, sizeof(header), 1, archive->fd) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't write: %s", strerror(errno));
		archive_free(archive);
		return -1;
	}

	archive_free(archive);

	return 0;
}

thdat_fptrs th07dat_fptrs = {
	th07_open,
	th07_write,
	th07_close
};
