/*
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that the
 * following conditions are met:
 *
 * 1. Redistributions of source code must retain this list
 *    of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce this
 *    list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include <config.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <errno.h>

#include "datpacker.h"
#include "thcrypt.h"
#include "thlzss.h"
#include "util.h"
#include "thdat.h"

static unsigned int
th95_get_crypt_param_index(const char *name)
{
	char index = 0;
	while(*name) index += *name++;
	return index & 7;
}

static const struct {
	unsigned char key;
	unsigned char step;
	unsigned int block;
	unsigned int limit;
} th95_crypt_params[] = {
	/* key  step  block  limit */
	{ 0x1b, 0x37, 0x40,  0x2800 },
	{ 0x51, 0xe9, 0x40,  0x3000 },
	{ 0xc1, 0x51, 0x80,  0x3200 },
	{ 0x03, 0x19, 0x400, 0x7800 },
	{ 0xab, 0xcd, 0x200, 0x2800 },
	{ 0x12, 0x34, 0x80,  0x3200 },
	{ 0x35, 0x97, 0x80,  0x2800 },
	{ 0x99, 0x37, 0x400, 0x2000 }
};

static archive_t*
th95_open(FILE* fd, unsigned int version)
{
	if (version != 10 && version != 11 && version != 95) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "internal error: wrong version passed: %u", version);
		return NULL;
	}

	return archive_open(fd, version, 16);
}

static int
th95_write(archive_t* archive, FILE* fd, const char* filename)
{
	unsigned int i;
	entry_t* entry;
	unsigned char* zdata;

	if (!archive) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "internal error: archive is NULL");
		return -1;
	}

	if (!filename) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "internal error: filename is NULL");
		archive_free(archive);
		return -1;
	}

	entry = thdat_write(archive, fd, filename);
	if (!entry) {
		archive_free(archive);
		return -1;
	}

	zdata = th_lz_fd(fd, &entry->zsize);
	if (!zdata) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "th_lz_fd failed (TODO: Improve error message.)");
		archive_free(archive);
		return -1;
	}
	if (entry->zsize >= entry->size) {
		if (fseek(fd, 0, SEEK_SET) == -1) {
			snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't seek: %s", strerror(errno));
			free(zdata);
			archive_free(archive);
			return -1;
		}

		if (fread(zdata, entry->size, 1, fd) != 1) {
			snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't read: %s", strerror(errno));
			free(zdata);
			archive_free(archive);
			return -1;
		}

		entry->zsize = entry->size;
	}

	i = th95_get_crypt_param_index(entry->name);
	if (th_encrypt(zdata, entry->zsize, th95_crypt_params[i].key, th95_crypt_params[i].step, th95_crypt_params[i].block, th95_crypt_params[i].limit) == -1) {
		free(zdata);
		archive_free(archive);
		return -1;
	}

	if (fwrite(zdata, entry->zsize, 1, archive->fd) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't write: %s", strerror(errno));
		free(zdata);
		archive_free(archive);
		return -1;
	}
	free(zdata);

	archive->offset += entry->zsize;

	return 0;
}

static int
th95_close(archive_t* archive)
{
	unsigned char* buffer;
	unsigned char* buffer_ptr;
	unsigned int i;
	unsigned char* zbuffer;
	uint32_t header[4];
	unsigned int list_size = 0;

	if (!archive) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "internal error: archive is NULL");
		return -1;
	}

	for (i = 0; i < archive->count; ++i) {
		const size_t namelen = strlen(archive->entries[i].name);
		list_size += (sizeof(uint32_t) * 3) + namelen + (4 - namelen % 4);
	}

	/* Write entry directory. */
	buffer = malloc(list_size);
	if (!buffer) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "malloc failed: %s", strerror(errno));
		archive_free(archive);
		return -1;
	}

	buffer_ptr = buffer;
	for (i = 0; i < archive->count; ++i) {
		const entry_t* entry = &archive->entries[i];
		const size_t namelen = strlen(entry->name);
		buffer_ptr = write_bytes((unsigned char*)entry->name, namelen + (4 - namelen % 4), buffer_ptr);
		buffer_ptr = write_uint32(entry->offset, buffer_ptr);
		buffer_ptr = write_uint32(entry->size, buffer_ptr);
		buffer_ptr = write_uint32(0, buffer_ptr);
	}

	unsigned int list_zsize = 0;
	zbuffer = th_lz_mem(buffer, list_size, &list_zsize);
	free(buffer);
	if (!zbuffer) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "th_lz_mem failed (TODO: Improve error message.)");
		archive_free(archive);
		return -1;
	}

	if (th_encrypt(zbuffer, list_zsize, 0x3e, 0x9b, 0x80, list_size) == -1) {
		free(zbuffer);
		archive_free(archive);
		return -1;
	}

	if (fwrite(zbuffer, list_zsize, 1, archive->fd) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't write: %s", strerror(errno));
		free(zbuffer);
		archive_free(archive);
		return -1;
	}
	free(zbuffer);

	if (fseek(archive->fd, 0, SEEK_SET) == -1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't seek: %s", strerror(errno));
		archive_free(archive);
		return -1;
	}

	/* Write header. */
	header[0] = 0x31414854;
	header[1] = list_size + 123456789;
	header[2] = list_zsize + 987654321;
	header[3] = archive->count - 0xf7e7f8ac;

	if (th_encrypt((unsigned char*)&header, sizeof(header), 0x1b, 0x37, sizeof(header), sizeof(header)) == -1) {
		archive_free(archive);
		return -1;
	}

	if (fwrite(&header, sizeof(header), 1, archive->fd) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't write: %s", strerror(errno));
		archive_free(archive);
		return -1;
	}

	archive_free(archive);

	return 0;
}

thdat_fptrs th95dat_fptrs = {
	th95_open,
	th95_write,
	th95_close
};
