/*
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that the
 * following conditions are met:
 *
 * 1. Redistributions of source code must retain this list
 *    of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce this
 *    list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include <config.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <libgen.h>

#include "datpacker.h"
#include "util.h"

archive_t*
archive_open(FILE* fd, uint32_t version, uint32_t offset)
{
	archive_t* archive;

	/* Write the header later. */
	if (fseek(fd, offset, SEEK_SET) == -1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't seek: %s", strerror(errno));
		return NULL;
	}

	archive = malloc(sizeof(archive_t));
	if (!archive) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "malloc failed: %s", strerror(errno));
		return NULL;
	}

	archive->version = version;
	archive->fd = fd;
	archive->offset = offset;
	archive->count = 0;
	archive->entries = NULL;

	return archive;
}

void
archive_free(archive_t* archive)
{
	if (archive) {
		free(archive->entries);
		free(archive);
	}
}

entry_t*
thdat_write(archive_t* archive, FILE* fd, const char* filename)
{
	int i;
	char* path;
	char* name;
	entry_t* e;

	archive->count++;
	archive->entries = realloc(archive->entries, archive->count * sizeof(entry_t));
	e = &archive->entries[archive->count - 1];

	memset(e->name, 0, 256);
	e->size = fsize(fd);
	e->zsize = 0;
	e->offset = archive->offset;
	e->extra = 0;

	if (e->size == -1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "fsize failed: %s", strerror(errno));
		return NULL;
	}

	path = strdup(filename);
	name = basename(path);

	strncpy(e->name, name, 255);

	/* check for duplicate entry */
	for (i = 0; i < archive->count - 1; ++i) {
		if (strcmp(name, archive->entries[i].name) == 0) {
			snprintf(library_error, LIBRARY_ERROR_SIZE, "duplicate filename %s", name);
			return NULL;
		}
	}

	free(path);

	return e;
}
